@@ -1,14 +1,15 @@
-/*
+/* vim:ts=4:ai
+ * 
  * File:	dwww-cache.c
  * Purpose:	Manage the dwww cache of converted documents.
  * Author:	Lars Wirzenius
- * Version:	"(#)dwww:$Id: dwww-cache.c,v 1.1.1.1 1996/12/12 06:41:08 jim Exp $"
+ * Version:	"(#)dwww:$Id: dwww-cache.c,v 1.12 2002/11/24 23:17:22 robert Exp $"
  * Description:	See the manual page for how to use this program.
  *
  *		Basically, what we do is read in a file from stdin,
  *		store it somewhere, and then retrieve it when asked.
- *		The original name of the file is given on the command 
- *		line.  We must be quick, so we keep a mapping between 
+ *		The original name of the file is given on the command
+ *		line.  We must be quick, so we keep a mapping between
  *		between the original name and the new name in a file.
  *
  *		The current implementation uses a simple "database"
@@ -24,6 +25,8 @@
  *		strings ("/usr/man/" for example).  We'll see.
  */
 
+
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -34,20 +37,37 @@
 #include <fcntl.h>
 #include <unistd.h>
 #include <errno.h>
+#include <time.h>
+#include <utime.h>
+#include <sys/file.h>
+#include <libgen.h>
+
+#include "utils.h"
+
+#define BUF_SIZE 1024
 
 /*
  * Filename prefixes (for simple compression). At most 256 elements!
  */
 static const char *prefs[] = {
 	"", 	/* MUST BE EMPTY STRING */
-	"/usr/man/man1/",
-	"/usr/man/man2/",
-	"/usr/man/man3/",
-	"/usr/man/man4/",
-	"/usr/man/man5/",
-	"/usr/man/man6/",
-	"/usr/man/man7/",
-	"/usr/man/man8/",
+	"/usr/lib/",
+	"/usr/share/doc/",
+	"/usr/doc/",
+	"/usr/local/doc/",
+	"/usr/share/info/",
+	"/usr/share/common-licenses/",
+	"/usr/info/",
+	"/usr/local/info/",
+	"/usr/share/man/man1/",
+	"/usr/share/man/man2/",
+	"/usr/share/man/man3/",
+	"/usr/share/man/man4/",
+	"/usr/share/man/man5/",
+	"/usr/share/man/man6/",
+	"/usr/share/man/man7/",
+	"/usr/share/man/man8/",
+	"/usr/share/",
 	"/usr/X11R6/man/man1/",
 	"/usr/X11R6/man/man2/",
 	"/usr/X11R6/man/man3/",
@@ -64,8 +84,14 @@ static const char *prefs[] = {
 	"/usr/local/man/man6/",
 	"/usr/local/man/man7/",
 	"/usr/local/man/man8/",
-	"/usr/doc/",
-	"/usr/local/doc/",
+	"/usr/man/man1/",
+	"/usr/man/man2/",
+	"/usr/man/man3/",
+	"/usr/man/man4/",
+	"/usr/man/man5/",
+	"/usr/man/man6/",
+	"/usr/man/man7/",
+	"/usr/man/man8/",
 };
 static int nprefs = sizeof(prefs) / sizeof(*prefs);
 
@@ -93,7 +119,7 @@ struct dochash {
 /*
  * Where we store the files and the database.
  */
-#define SPOOL_DIR "/var/spool/dwww/"
+#define SPOOL_DIR "/var/cache/dwww/"
 
 /*
  * The name of the database in SPOOL_DIR.
@@ -137,7 +163,7 @@ static int list(char *, char *);
 static int list_all(char *, char *);
 static int clean(char *, char *);
 
-static int help(char *type, char *location) { puts("help"); return 0; }
+static int help(char *type, char *location) { (void) puts("help"); return 0; }
 
 static int open_db_reading(void);
 static int open_db_writing(void);
@@ -151,15 +177,15 @@ static int close_db(int);
 static int storage_size(struct cache_entry *);
 static int output_file(char *);
 static void init_entry(struct cache_entry *, char *, char *, char *, int);
-static int create_new_file(char *, char *, char *);
-static int copy_fd_to_fd(int, int);
+static int create_tmp_file(char *, char *, size_t);
+static int get_new_filename(char *, char *, char *, size_t);
+static int copy_fd_to_fd(int, int, int);
 static char *int_to_type(int);
 static int type_to_int(char *);
 static int rebuild_db(struct dochash *);
-static int file_exists(char *);
-static long mtime(char *);
+static time_t mtime(char *);
 static int find_pref(const char *);
-
+static int set_modtime(const char *, time_t); 
 
 int main(int argc, char **argv) {
 	static struct {
@@ -177,7 +203,8 @@ int main(int argc, char **argv) {
 	};
 	int i, first_nonopt;
 
-	set_progname(argv[0], "dwww-cache");
+	dwww_initialize( "dwww-cache");
+
 
 	if (parse_options(argc, argv, options, &first_nonopt) == -1)
 		return EXIT_FAILURE;
@@ -220,8 +247,8 @@ static int list(char *type, char *location) {
 	struct dochash hash;
 	struct cache_entry *data;
 	int i, db;
-	char buf[1024];
-	char orig[1024];
+	char buf[BUF_SIZE];
+	char orig[BUF_SIZE];
 
 	db = open_db_reading();
 	if (db == -1)
@@ -235,11 +262,13 @@ static int list(char *type, char *location) {
 #endif
 		return -1;
 	}
+	
 	data = hash.tab.data;
-	sprintf(orig, "%s%s", prefs[data[i].pref], data[i].original);
-	sprintf(buf, "%s%s", SPOOL_DIR, data[i].converted);
-	if (mtime(orig) > mtime(buf)) {
+	snprintf(orig, sizeof(orig), "%s%s", prefs[data[i].pref], data[i].original);
+	snprintf(buf,  sizeof(buf),  "%s%.20s", SPOOL_DIR, data[i].converted);
+	if (mtime(orig) != mtime(buf)) {
 		(void) close_db(db);
+		(void) truncate(buf, 0);
 		return -1;
 	}
 	printf("%s %s %s %s\n", int_to_type(data[i].type),
@@ -256,8 +285,8 @@ static int list_all(char *type, char *location) {
 	struct dochash hash;
 	struct cache_entry *data;
 	int i, db;
-	char buf[1024];
-	char orig[1024];
+	char buf[BUF_SIZE];
+	char orig[BUF_SIZE];
 
 	db = open_db_reading();
 	if (db == -1)
@@ -265,16 +294,21 @@ static int list_all(char *type, char *location) {
 	if (read_db(db, &hash) == -1)
 		return -1;
 	data = hash.tab.data;
+	
 	for (i = 0; i < hash.tab.used; ++i) {
-		sprintf(orig, "%s%s", prefs[data[i].pref], data[i].original);
-		sprintf(buf, "%s%s", SPOOL_DIR, data[i].converted);
-		if (mtime(orig) <= mtime(buf))
+		snprintf(orig, sizeof(orig), "%s%s", prefs[data[i].pref], data[i].original);
+		snprintf(buf,  sizeof(buf),  "%s%.20s", SPOOL_DIR, data[i].converted);
+		if (mtime(orig) != mtime(buf)) 
+			(void) truncate(buf, 0);
+		else
 			printf("%s %s %s %s\n", int_to_type(data[i].type),
 				orig, data[i].converted,
 				data[i].permanent ? "y" : "n");
 	}
+	
 	if (close_db(db) == -1)
 		return -1;
+	
 	return 0;
 }
 
@@ -284,8 +318,8 @@ static int lookup(char *type, char *location) {
 	struct dochash hash;
 	struct cache_entry *data;
 	int i, db;
-	char buf[1024];
-	char orig[1024];
+	char buf[BUF_SIZE];
+	char orig[BUF_SIZE];
 
 	db = open_db_reading();
 	if (db == -1)
@@ -296,14 +330,18 @@ static int lookup(char *type, char *location) {
 	if (i == -1)
 		return -1;
 	data = hash.tab.data;
-	sprintf(orig, "%s%s", prefs[data[i].pref], data[i].original);
-	sprintf(buf, "%s%s", SPOOL_DIR, data[i].converted);
-	if (mtime(orig) > mtime(buf)) {
+	snprintf(orig, sizeof(orig), "%s%s", prefs[data[i].pref], data[i].original);
+	snprintf(buf,  sizeof(buf),  "%s%.20s", SPOOL_DIR, data[i].converted);
+	if (mtime(orig) != mtime(buf)) {
+		(void) close_db(db);
+		(void) truncate(buf, 0);
+		return -1;
+	}
+	if (output_file(buf) == -1) {
 		(void) close_db(db);
 		return -1;
 	}
-	if (output_file(buf) == -1 ||
-	    close_db(db) == -1)
+	if (close_db(db) == -1)
 		return -1;
 	return 0;
 }
@@ -314,48 +352,80 @@ static int store(char *type, char *location) {
 	int i, db, fd;
 	struct dochash hash;
 	struct cache_entry new, *data;
-	char oldname[1024];
-	char buf[1024];
-	char full[1024];
+	char filename[BUF_SIZE];
+	char buf[BUF_SIZE];
+	char tmp_file[BUF_SIZE];
 
-	fd = create_new_file(SPOOL_DIR, "", buf);
+	fd = create_tmp_file(SPOOL_DIR, tmp_file, sizeof(tmp_file));
 	if (fd == -1) {
-		errormsg(0, -1, "can't create cache file");
-		return -1;
+		errormsg(0, -1, "can't create temporary cache file");
+		fd = STDOUT_FILENO;
 	}
-	sprintf(full, "%s/%s", SPOOL_DIR, buf);
 	
-	if (copy_fd_to_fd(0, fd) == -1) {
-		errormsg(0, -1, "can't copy stdin to %s", full);
-		(void) unlink(full);
+	if (copy_fd_to_fd(STDIN_FILENO, fd, STDOUT_FILENO) == -1) {
+		errormsg(0, -1, "can't copy stdin to %s", tmp_file);
+		(void) unlink(tmp_file);
 		return -1;
 	}
 	
+	if (fd == STDOUT_FILENO)
+		return -1;
+	
 	close(fd);
 
 	db = open_db_writing();
 	if (db == -1) {
-		(void) unlink(full);
+		(void) unlink(tmp_file);
 		return -1;
 	}
 		
-	init_entry(&new, type, location, buf, 0);
 
-	if (read_db(db, &hash) == -1)
+	if (read_db(db, &hash) == -1) {
+		(void) close_db(db);
+		(void) unlink (tmp_file);
 		return -1;
+	}	
 		
 	i = lookup_db(&hash, location);
 	if (i != -1) {
 		data =  hash.tab.data;
-		sprintf(oldname, "%s%s", SPOOL_DIR, data[i].converted);
-		(void) unlink(oldname);
+		snprintf(filename, sizeof(filename), "%s%s", SPOOL_DIR, data[i].converted);
+		(void) unlink(filename);
+	} else {
+		if (get_new_filename(location, SPOOL_DIR, buf, sizeof(buf)) == -1) {
+			errormsg(0, -1, "can't get new cache filename");
+			(void) unlink(tmp_file);
+			(void) close_db(db);
+			return -1;
+		}
+		snprintf(filename, sizeof(filename), "%s%s", SPOOL_DIR, buf);
+		
+		init_entry(&new, type, location, buf, 0);
+
+		if (insert_db(&hash, &new) == -1)
+		{ 
+			(void) unlink (tmp_file);
+			(void) close_db(db);
+			return -1;
+		}
 	}
 
-	if (insert_db(&hash, &new) == -1 ||
-	    write_db(db, &hash) == -1 ||
-	    close_db(db) == -1)
-		return -1;
+	if (rename(tmp_file, filename) == -1)
+	{
+			errormsg(0, -1, "can't rename temporary file to %s", filename);
+			(void) unlink(tmp_file);
+			(void) close_db(db);
+			return -1;
+	}
 
+	(void) set_modtime(filename, mtime(location));
+	
+	if ((write_db(db, &hash) == -1)  || (close_db(db) == -1)){
+			(void) truncate(filename, 0);
+			(void) close_db(db);
+			return -1;
+	}
+			
 	return 0;
 }
 
@@ -364,8 +434,8 @@ static int store(char *type, char *location) {
 static int clean(char *type, char *location) {
 	struct dochash hash;
 	struct cache_entry *data;
-	char buf[1024];
-	char orig[1024];
+	char buf[BUF_SIZE];
+	char orig[BUF_SIZE];
 	int i, j, db;
 
 	db = open_db_writing();
@@ -375,9 +445,9 @@ static int clean(char *type, char *location) {
 	data = hash.tab.data;
 	j = 0;
 	for (i = 0; i < hash.tab.used; ++i) {
-		sprintf(orig, "%s%s", prefs[data[i].pref], data[i].original);
-		sprintf(buf, "%s%s", SPOOL_DIR, data[i].converted);
-		if (file_exists(orig) && file_exists(buf))
+		snprintf(orig, sizeof(orig), "%s%s", prefs[data[i].pref], data[i].original);
+		snprintf(buf,  sizeof(buf),  "%s%.20s", SPOOL_DIR, data[i].converted);
+		if (mtime(orig) == mtime(buf)) 
 			data[j++] = data[i];
 		else
 			(void) unlink(buf);
@@ -396,25 +466,49 @@ static int clean(char *type, char *location) {
 
 static int open_db_reading(void) {
 	int db;
+	int tries = 5;
 
-	db = open(CACHE_DB, O_RDONLY, 0);
-	if (db == -1 && errno != ENOENT)
-		errormsg(0, -1, "can't check cache database");
-	
-	/* fixme: create shared lock here */	
-	
+	db = open(CACHE_DB, O_RDONLY, 0644);
+	if (db == -1) {
+			if (errno != ENOENT)
+				errormsg(0, -1, "can't check cache database");
+		return -1;
+	}				
+
+	while (--tries && flock(db, LOCK_SH | LOCK_NB) < 0) 
+			sleep(1);
+
+	if (!tries)
+	{
+		(void) close(db);
+		errormsg(0, -1, "can't create shared lock on cache database");
+		return -1;
+	}
+		
 	return db;
 }
 
 
 static int open_db_writing(void) {
 	int db;
+	int tries = 5;
 
 	db = open(CACHE_DB, O_RDWR | O_CREAT, 0664);
 	if (db == -1)
+	{
 		errormsg(0, -1, "can't update cache database");
+		return -1;
+	}		
+					
+	while (--tries && flock(db, LOCK_EX | LOCK_NB) < 0) 
+			sleep(1);
 
-	/* fixme: lock the database exclusively */
+	if (!tries)
+	{
+		(void) close(db);
+		errormsg(0, -1, "can't create exlusive lock on cache database");
+		return -1;
+	}
 
 	return db;
 }
@@ -430,13 +524,13 @@ static int read_db(int db, struct dochash *hash) {
 		return -1;
 	}
 
-	p = malloc(st.st_size);
+	p = malloc((size_t)st.st_size);
 	if (p == NULL) {
 		errormsg(0, -1, "out of memory");
 		return -1;
 	}
 	
-	if (read(db, p, st.st_size) != st.st_size) {
+	if (read(db, p, (size_t)st.st_size) != st.st_size) {
 		errormsg(0, -1, "couldn't read database");
 		return -1;
 	}
@@ -492,7 +586,7 @@ static int lookup_db(struct dochash *hash, char *original) {
 
 	init_entry(&dummy, "", original, 0, 0);
 	data = hash->tab.data;
-	p = bsearch(&dummy, data, hash->tab.used, 
+	p = bsearch(&dummy, data, hash->tab.used,
 		sizeof(struct cache_entry), compare_entry);
 	if (p == NULL)
 		return -1;
@@ -565,7 +659,7 @@ static int write_db(int db, struct dochash *hash) {
 	}
 	
 	if (ftruncate(db, 0) == -1) {
-		errormsg(0, -1, "can't update truncate database");
+		errormsg(0, -1, "can't truncate cache database");
 		return -1;
 	}
 	
@@ -579,7 +673,11 @@ static int write_db(int db, struct dochash *hash) {
 
 
 static int close_db(int db) {
-	/* fixme: release locks */
+	if (flock(db, LOCK_UN) == -1) {
+		close(db);
+		errormsg(0, -1, "can't unlock cache database");
+		return -1;
+	}
 	if (close(db) == -1) {
 		errormsg(0, -1, "error closing database");
 		return -1;
@@ -614,8 +712,8 @@ static int output_file(char *name) {
 		return -1;
 	}
 	
-	if (copy_fd_to_fd(fd, 1) == -1) {
-		close(fd);
+	if (copy_fd_to_fd(fd, 1, 1) == -1) {
+		(void) close(fd);
 		return -1;
 	}
 	
@@ -624,17 +722,34 @@ static int output_file(char *name) {
 }
 
 
-static int copy_fd_to_fd(int from, int to) {
-	char buf[1024];
-	int n;
+static int copy_fd_to_fd(int from, int to_prim, int to_sec) {
+	char buf[BUF_SIZE];
+	size_t n;
+	ssize_t n_prim = 0;
+	ssize_t n_sec  = 0;
+
+	
+	if (to_prim == to_sec)
+		n_sec = -100; /* *must* be less then 0 */
 	
 	while ((n = read(from, buf, sizeof(buf))) > 0) {
-		if (write(to, buf, n) != n) {
+		if (n_prim >= 0)
+			n_prim = write(to_prim, buf, n) - n;
+		if (n_sec >= 0)
+			n_sec = write(to_sec, buf, n) - n;
+		if (n_prim < 0 && n_sec < 0)
+		{	
 			errormsg(0, -1, "error writing");
 			return -1;
 		}
 	}
 	
+	if (n_prim < 0)
+	{	
+		errormsg(0, -1, "error writing");
+		return -1;
+	}
+	
 	if (n < 0) {
 		errormsg(0, -1, "error reading");
 		return -1;
@@ -644,22 +759,47 @@ static int copy_fd_to_fd(int from, int to) {
 }
 
 
-static int create_new_file(char *prefix, char *suffix, char *full) {
-	char buf[1024];
-	int i, fd;
+static int create_tmp_file(char * dir, char * full, size_t size) {
+	int fd;
 
-	i = 0;
-	do {
-		sprintf(buf, "%s%d%s", prefix, i, suffix);
-		fd = open(buf, O_CREAT | O_EXCL | O_RDWR, 0664);
-	} while (fd == -1 && errno == EEXIST && i++ < INT_MAX-1);
-
-	if (fd != -1)
-		sprintf(full, "%d%s", i, suffix);
+	snprintf(full, size,  "%s/tmp%d.%ld", dir, (int) getpid(), (long int)time(NULL));
+	fd = open(full, O_CREAT | O_EXCL | O_RDWR, 0664);
+	
 	return fd;
 }
 
+static int get_new_filename(char * location, char * dir, char * buf, size_t buf_size)
+{
+		char * tmp;
+		char full[BUF_SIZE];
+		int  i = 0;
+		char c = '_';
 
+		tmp = basename(location);
+		if (tmp != NULL) 
+				c = *tmp;
+		
+		snprintf(full, sizeof(full), "%s/%c", dir, c);
+		if (mkdir(full, 0775) == -1 && errno != EEXIST)
+		{
+				errormsg(0, -1, "can't create directory %d", full);
+				return -1;
+		}
+		for (i = 0; i < INT_MAX; i++)
+		{
+				snprintf(full, sizeof(full), "%s/%c/%d", dir, c, i);
+				errno = 0;
+				if (access(full, F_OK) == -1 && errno == ENOENT)
+				{
+						snprintf(buf, buf_size, "%c/%d", c, i);
+						return 0;
+				}
+				else if (errno)
+						return -1;
+		}
+		return -1;
+}
+		
 
 static struct {
 	char *str;
@@ -670,6 +810,9 @@ static struct {
 	{ "info", 'i' },
 	{ "text", 't' },
 	{ "dir", 'd' },
+	{ "html", 'h' },
+	{ "text/html", 'h' },
+	{ "file", 'f' },
 	{ NULL, '?' },
 };
 
@@ -694,16 +837,12 @@ static char *int_to_type(int ch) {
 }
 
 
-static int file_exists(char *name) {
-	struct stat st;
-	return stat(name, &st) != -1;
-}
-
-
-static long mtime(char *name) {
+static time_t mtime(char *name) {
 	struct stat st;
 	if (stat(name, &st) == -1)
 		return 0;
+	if (!st.st_size) /* file is empty ??? */
+		return 0;
 	return st.st_mtime;
 }
 
@@ -715,3 +854,14 @@ static int find_pref(const char *s) {
 			return i;
 	return 0;
 }
+
+
+static int set_modtime(const char * file, time_t modtime) {
+	struct utimbuf ut;
+
+	ut.modtime = modtime;
+	ut.actime  = time(NULL);
+
+	return utime(file, &ut); 			
+}
+
